//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// NesMapper21.cpp - Implementation of mapper #23 (Konami VRC4)
//////////////////////////////////////////////////////////////////////

#include "NesMapper021.h"



NesMapper21::NesMapper21(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *&linkPPU)
{
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;
	
	// Replace the standard PPU memory object with a VRC4 specific one	
	NesPPU *newPPU = new VRC4PPU();
	newPPU->Init();
	newPPU->VRAM->mbytMirroringType = linkPPU->VRAM->mbytMirroringType;
	newPPU->VRAM->link(this);	
	delete mPpu;
	mPpu = newPPU;
	linkPPU = mPpu;
	mCpu->linkCPU(mPpu, mCpu->mApu, mCpu->mController1, mCpu->mController2, this);

	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	
	for (int cbanks = 0; cbanks < 8; cbanks++) { chrbanks[cbanks] = 0; }
	initMaps();
	intSwitchSpace = 0x8000;
	intEnableLatch = 0;
	intIrqCount = 0;
}


void NesMapper21::initMemory() {	
	if (bytPRGCount > 0) {
		swapPRGBank8(0x8000, 0);			  // First 16k bank to $8000
		swapPRGBank8(0xA000, 1);			  // First 16k bank to $8000

		swapPRGBank8(0xC000, (bytPRGCount << 1) - 2); // Last 16k bank to $C000
		swapPRGBank8(0xE000, (bytPRGCount << 1) - 1); // Last 16k bank to $C000
	}
}


void NesMapper21::write(Uint16 addr, unsigned char data)
{	
	switch (addr) {
	case 0x8000:	// Select 8K PRGROM page at $8000
		swapPRGBank8(intSwitchSpace, (data % (bytPRGCount << 1)) );
		break;

	case 0x9000:	// Mirorring select
		switch (data & 0x3) {
		case 0:
			mPpu->VRAM->mbytMirroringType = VERTICAL;
			break;
		case 1:
			mPpu->VRAM->mbytMirroringType = HORIZONTAL;
			break;
		case 2:
			mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
			mPpu->VRAM->mintSSMirrorLocation = 0x2000;
			break;
		case 3:
			mPpu->VRAM->mbytMirroringType = SINGLESCREEN;
			mPpu->VRAM->mintSSMirrorLocation = 0x2400;
			break;
		}
		break;
	
	case 0x9002:	// SaveRAM toggle, switchable space select
		intSwitchSpace = (data & 0x02) ? 0xC000 : 0x8000;
		break;

	case 0xA000:	// Select 8K PRGROM bank at $A000
		swapPRGBank8(0xA000, (data % (bytPRGCount << 1)) );
		break;
	
	// Low and high nibbles of 1K CHRROM at PPU $0000
	case 0xB000: // low 4 bits
		chrbanks[0] = (chrbanks[0] & 0xF0) | (data & 0x0F);
		swapCHRBank1(0x0000, chrbanks[0] % (bytCHRCount << 3));
		break;
	case 0xB002: // high 4 bits
		chrbanks[0] = (chrbanks[0] & 0x0F) | ((data & 0x0F) << 4);
		swapCHRBank1(0x0000, chrbanks[0] % (bytCHRCount << 3));
		break;

	// Low and high nibbles of 1K CHRROM at PPU $0400
	case 0xB004:
	case 0xB001: // low 4 bits
		chrbanks[1] = (chrbanks[1] & 0xF0) | (data & 0x0F);
		swapCHRBank1(0x0400, chrbanks[1] % (bytCHRCount << 3));
		break;
	case 0xB006:
	case 0xB003: // high 4 bits
		chrbanks[1] = (chrbanks[1] & 0x0F) | ((data & 0x0F) << 4);
		swapCHRBank1(0x0400, chrbanks[1] % (bytCHRCount << 3));
		break;	
	
	// Low and high nibbles of 1K CHRROM at PPU $0800
	case 0xC000: // low 4 bits
		chrbanks[2] = (chrbanks[2] & 0xF0) | (data & 0x0F);
		swapCHRBank1(0x0800, chrbanks[2] % (bytCHRCount << 3));
		break;
	case 0xC002: // high 4 bits
		chrbanks[2] = (chrbanks[2] & 0x0F) | ((data & 0x0F) << 4);
		swapCHRBank1(0x0800, chrbanks[2] % (bytCHRCount << 3));
		break;

	// Low and high nibbles of 1K CHRROM at PPU $0C00
	case 0xC004:
	case 0xC001: // low 4 bits
		chrbanks[3] = (chrbanks[3] & 0xF0) | (data & 0x0F);
		swapCHRBank1(0x0C00, chrbanks[3] % (bytCHRCount << 3));
		break;
	case 0xC006:
	case 0xC003: // high 4 bits
		chrbanks[3] = (chrbanks[3] & 0x0F) | ((data & 0x0F) << 4);
		swapCHRBank1(0x0C00, chrbanks[3] % (bytCHRCount << 3));
		break;
	
		
	// Low and high nibbles of 1K CHRROM at PPU $1000
	case 0xD000: // low 4 bits
		chrbanks[4] = (chrbanks[4] & 0xF0) | (data & 0x0F);
		swapCHRBank1(0x1000, chrbanks[4] % (bytCHRCount << 3));
		break;
	case 0xD002: // high 4 bits
		chrbanks[4] = (chrbanks[4] & 0x0F) | ((data & 0x0F) << 4);
		swapCHRBank1(0x1000, chrbanks[4] % (bytCHRCount << 3));
		break;

	// Low and high nibbles of 1K CHRROM at PPU $1400
	case 0xD004:
	case 0xD001: // low 4 bits
		chrbanks[5] = (chrbanks[5] & 0xF0) | (data & 0x0F);
		swapCHRBank1(0x1400, chrbanks[5] % (bytCHRCount << 3));
		break;
	case 0xD006:
	case 0xD003: // high 4 bits
		chrbanks[5] = (chrbanks[5] & 0x0F) | ((data & 0x0F) << 4);
		swapCHRBank1(0x1400, chrbanks[5] % (bytCHRCount << 3));
		break;	
	
	// Low and high nibbles of 1K CHRROM at PPU $1800
	case 0xE000: // low 4 bits
		chrbanks[6] = (chrbanks[6] & 0xF0) | (data & 0x0F);
		swapCHRBank1(0x1800, chrbanks[6] % (bytCHRCount << 3));
		break;
	case 0xE002: // high 4 bits
		chrbanks[6] = (chrbanks[6] & 0x0F) | ((data & 0x0F) << 4);
		swapCHRBank1(0x1800, chrbanks[6] % (bytCHRCount << 3));
		break;

	// Low and high nibbles of 1K CHRROM at PPU $1C00
	case 0xE004:
	case 0xE001: // low 4 bits
		chrbanks[7] = (chrbanks[7] & 0xF0) | (data & 0x0F);
		swapCHRBank1(0x1C00, chrbanks[7] % (bytCHRCount << 3));
		break;
	case 0xE006:
	case 0xE003: // high 4 bits
		chrbanks[7] = (chrbanks[7] & 0x0F) | ((data & 0x0F) << 4);
		swapCHRBank1(0x1C00, chrbanks[7] % (bytCHRCount << 3));
		break;	

	case 0xF000:		// IRQ Counter register	
		intIrqLatch = (intIrqLatch & 0xF0) | (data & 0x0F);
		break;

	case 0xF002:		// IRQ Control Register #0		
		intIrqLatch = (intIrqLatch & 0x0F) | ((data & 0x0F)  << 4);
		break;

	case 0xF003:		// IRQ Control Register #1
		// Reset counter to zero
		blnIrqEnable = (intEnableLatch == 0x01) ? true : false;
		break;
	
	case 0xF004:		// ?
		intIrqCount = intIrqLatch;
		blnIrqEnable = (data & 0x02) ? true : false;
		intEnableLatch = data & 0x01;
		break;
	}
}