//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// NesMapper064.cpp - Implementation of mapper #64 (Tengen RAMBO-1)
//////////////////////////////////////////////////////////////////////

#include "NesMapper064.h"


NesMapper64::NesMapper64(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *&linkPPU) {
	mCartridge = linkCartridge;
	mCpu = linkCPU;
	mPpu = linkPPU;

	// Replace the standard PPU memory object with a VRC4 specific one	
	NesPPU *newPPU = new TengenPPU();
	newPPU->Init();
	newPPU->VRAM->mbytMirroringType = linkPPU->VRAM->mbytMirroringType;
	newPPU->VRAM->link(this);	
	delete mPpu;
	mPpu = newPPU;
	linkPPU = mPpu;
	mCpu->linkCPU(mPpu, mCpu->mApu, mCpu->mController1, mCpu->mController2, this);


	bytPRGCount = mCartridge->header.PRGROM_count;
	bytCHRCount = mCartridge->header.CHRROM_count;
	
	initMaps();	
}


void NesMapper64::initMemory() {	
	if (bytPRGCount > 0) {
		// Last 8K PRGROM goes to all 4 slots
		swapPRGBank8(0x8000, (bytPRGCount << 1) - 1);
		swapPRGBank8(0xA000, (bytPRGCount << 1) - 1);
		swapPRGBank8(0xC000, (bytPRGCount << 1) - 1);
		swapPRGBank8(0xE000, (bytPRGCount << 1) - 1);
	}
	
	if (bytCHRCount) swapCHRBank8(0); // First 8K of CHRROM if exists

	// Reset these guys
	reg0 = 0;
	blnIrqEnable = false;
	intIrqCount = intIrqLatch = intIrqMode = intIrqNext = intIrqReset = 0;
}


void NesMapper64::write(Uint16 addr, unsigned char data) {		
	switch (addr & 0xF003) {
	case 0x8000:	// Set the command
		reg0 = data;
		break;

	case 0x8001:	// Activate the command
		switch (reg0 & 0x0F) {
		case 0:		// Select 2 1K VROM pages at $0000			
			swapCHRBank2(0x0000 ^ ((reg0 & 0x80) ? 0x1000 : 0x0000), ( (data >> 1) % (bytCHRCount << 2)) );
			break;
		case 1:		// Select 2 1K VROM pages at $0800
			swapCHRBank2(0x0800 ^ ((reg0 & 0x80) ? 0x1000 : 0x0000), ( (data >> 1) % (bytCHRCount << 2)) );
			break;
		case 2:		// Select 1K VROM page at $1000			
			//swapCHRBank1(0x1000 ^ ((reg0 & 0x80) ? 0x1000 : 0x0000), (data % bytCHRCount << 3) );
			swapCHRBank1(0x1000 ^ ((reg0 & 0x80) ? 0x1000 : 0x0000), (data % (bytCHRCount << 3)) );
			break;
		case 3:		// Select 1K VROM page at $1400
			swapCHRBank1(0x1400 ^ ((reg0 & 0x80) ? 0x1000 : 0x0000), (data % (bytCHRCount << 3)) );
			break;
		case 4:		// Select 1K VROM page at $1800
			swapCHRBank1(0x1800 ^ ((reg0 & 0x80) ? 0x1000 : 0x0000), (data % (bytCHRCount << 3)) );
			break;
		case 5:		// Select 1K VROM page at $1C00
			swapCHRBank1(0x1C00 ^ ((reg0 & 0x80) ? 0x1000 : 0x0000), (data % (bytCHRCount << 3)) );
			break;
		
		case 6:		// Select first switchable ROM page
			swapPRGBank8( (reg0 & 0x40) ? 0xA000 : 0x8000, data % (bytPRGCount << 1) );
			break;
		
		case 7:		// Select second switchable ROM page
			swapPRGBank8( (reg0 & 0x40) ? 0xC000 : 0xA000, data % (bytPRGCount << 1) );
			break;
		
		case 8:		// Select 1K VROM page at PPU $0400
			if (data & 0x20) {
				swapCHRBank1(0x0400, (data % (bytCHRCount << 3)) );
			}
			break;
		
		case 9:		// Select 1K VROM page at PPU $0800
			swapCHRBank1(0x0C00, (data % (bytCHRCount << 3)) );			
			break;
		
		case 15:	// Select third switchable ROM page
			swapPRGBank8( (reg0 & 0x40) ? 0x8000 : 0xC000, data % (bytPRGCount << 1) );
			break;
		}
		break;

	case 0xA000:	// Set mirroring
		mPpu->VRAM->mbytMirroringType = (data & 0x01) ? HORIZONTAL : VERTICAL;
		break;
	
	case 0xC000:	// Set latch, possibly load IRQ counter
		intIrqLatch = data;
		if (intIrqReset) intIrqCount = intIrqLatch;
		break;

	case 0xC001:
		intIrqReset = 1;
		intIrqCount = intIrqLatch;
		intIrqNext = 0;
		intIrqMode = data & 0x01;
		break;

	case 0xE000:	// Disable IRQs
		blnIrqEnable = false;
		if (intIrqReset) intIrqCount = intIrqLatch;
		break;

	case 0xE001:	// Enable IRQs
		blnIrqEnable = true;
		if (intIrqReset) intIrqCount = intIrqLatch;
		break;

	}
}




//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the mapper to the file stream
//-------------------------------------------------------------------
void NesMapper64::saveState(FILE *stream) {
	NesBaseMapper::saveState(stream);
	fwrite(&blnIrqEnable, sizeof(blnIrqEnable), 1, stream);
	fwrite(&intIrqCount, sizeof(intIrqCount), 1, stream);
	fwrite(&intIrqLatch, sizeof(intIrqLatch), 1, stream);
	fwrite(&intIrqMode, sizeof(intIrqMode), 1, stream);
	fwrite(&reg0, sizeof(reg0), 1, stream);
}





//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the mapper from the file stream
//-------------------------------------------------------------------
void NesMapper64::loadState(FILE *stream) {
	NesBaseMapper::loadState(stream);	
	fread(&blnIrqEnable, sizeof(blnIrqEnable), 1, stream);
	fread(&intIrqCount, sizeof(intIrqCount), 1, stream);
	fread(&intIrqLatch, sizeof(intIrqLatch), 1, stream);
	fread(&intIrqMode, sizeof(intIrqMode), 1, stream);
	fread(&reg0, sizeof(reg0), 1, stream);
}