//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// NesBaseMapper.h: interface for the NesBaseMapper class.
//////////////////////////////////////////////////////////////////////
// Class Description:  The NesBaseMapper class is responsible for
// transferring data from a NesCartridge object to the NES CPU memory
// and the NES PPU memory (mapper #0).  It also provides the base class 
// from which specialized mappers can be derived.  Specialized mappers
// often accept commands from the CPU via the "write" procedure, and these
// special commands often instruct the Mapper to swap various banks of data 
// from the cartridge into the CPU/PPU memory.
//
// This class also provides some commonly-used bank swapping procedures
// that are used by more advanced  (derived) classes.
//
// The PrgMap8k PRGROM map is always maintained and up to date.  (8K is the
// smallest swappable unit of data for the PRGROM)
//////////////////////////////////////////////////////////////////////


#ifndef NES_BASE_MAPPER
#define NES_BASE_MAPPER

#include "Porting/basetypes.h"

#include "NesCartridge.h"
#include "NesCPU.h"
#include "NesPPU.h"


class NesBaseMapper
{
public:
	///////////////////////////////
	// Class Procedures
	///////////////////////////////
	NesBaseMapper() { };
	NesBaseMapper(NesCartridge *linkCartridge, NesCPU *linkCPU, NesPPU *linkPPU);
	virtual ~NesBaseMapper() { };
	void initMaps();
	virtual void initMemory();
	virtual void write(Uint16 addr, Uint8 data);
	virtual Uint8 read(Uint16 addr) { return mCpu->memory[addr]; }

	
	// Provide swapping routines for more complex mappers	
	void swapPRGBank8(Uint16 intOffset, Uint8 bytBank);
	void swapPRGBank16(Uint16 intOffset, Uint8 bytBank);
	void swapPRGBank32(Uint8 bytBank);
	
	void swapCHRBank1(Uint16 intOffset, Uint8 bytBank);
	void swapCHRBank2(Uint16 intOffset, Uint8 bytBank);
	void swapCHRBank4(Uint16 intOffset, Uint8 bytBank);
	void swapCHRBank8(Uint8 bytBank);

	virtual void saveState(FILE *stream);
	virtual void loadState(FILE *stream);

	
	///////////////////////////////
	// Class Data Members
	///////////////////////////////
	// Need access to following system components: Cartridge, CPU, PPU
	NesCartridge *mCartridge;
	NesCPU *mCpu;
	NesPPU *mPpu;

	// For convenience, mappers also have their own local
	// copies of the PRG and CHR ROM count	
	unsigned char bytPRGCount, bytCHRCount;
	
	/////// These array map the current banks held in the memory, and prevent unneeded swapping ///////
	int ChrMap8k[1];
	int ChrMap4k[2];
	int ChrMap2k[4];
	int ChrMap1k[8];

	int PrgMap32k[1];
	int PrgMap16k[2];
	int PrgMap8k[4];
};

#endif 
