//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// NesCartridge.cpp: implementation of the NesCartridge class.
//////////////////////////////////////////////////////////////////////

#include "NesCartridge.h"

//////////////////////////////////////////////////////////////////////
// Procedure: NesCartridge Constructor
//////////////////////////////////////////////////////////////////////
// Description: The constructor for the NesCartridge accepts 1 parameter.
// This is a buffer that contains the contents of a NES file.  So in order
// to create a NesCartridge object, you open a NES file, read it's contents
// into memory, and instantiate a NesCartridge with a pointer to this memory.
// See Nes::loadCartridge for an example of this.
//
// A standard constructor is also provided that sets the data members to
// safe values
//////////////////////////////////////////////////////////////////////
NesCartridge::NesCartridge() {
	PRGROM_Pages = NULL;
	CHRROM_Pages = NULL;
	header.PRGROM_count = header.CHRROM_count = 0;
}



NesCartridge::NesCartridge(unsigned char *&cartdata) {
	mintErrorCode = OK;
	header.PRGROM_count = 0;
	header.CHRROM_count = 0;

	PRGROM_Pages = NULL;
	CHRROM_Pages = NULL;
				
	header.name[0] = (char) cartdata[0];
	header.name[1] = (char) cartdata[1];
	header.name[2] = (char) cartdata[2];
	header.name[3] = '\0';
		
	if (strcmp(header.name, "NES") != 0) {
		mintErrorCode = INVALID_HEADER;			
	} else {
		header.code = cartdata[3];
		header.PRGROM_count = cartdata[4];
		header.CHRROM_count = cartdata[5];
		header.ROM_Control_1 = cartdata[6];
		header.ROM_Control_2 = cartdata[7];
			
		unsigned char *garbage = &cartdata[8];
						
		// Check for a particularly common header error
		if ((garbage[0] == 'i') && (garbage[1] == 's')) {
			// Bad header, control byte 2 is trash
			header.mapper = ((header.ROM_Control_1 & 0xF0) >> 4);
		} else {
			// Ok header (we think), control byte 2 is ok
			header.mapper = ((header.ROM_Control_1 & 0xF0) >> 4) | (header.ROM_Control_2 & 0xF0);
		}
								
		if (header.PRGROM_count > 0) {
			// Allocate an int pointer for each pair of prgrom pages
			PRGROM_Pages = new Uint8 *[header.PRGROM_count];		
			int intPageCtr = 0;
									
			// Load all pages of PRGROM into memory
			while (intPageCtr < header.PRGROM_count) {				
				// Allocate storage for this pair of pages
				PRGROM_Pages[intPageCtr] = new Uint8[16384];

				// Load each page into memory.								
				memcpy(&PRGROM_Pages[intPageCtr][0], &cartdata[16 + intPageCtr*16384], 16384);
													
				intPageCtr++;		// next 16k page
			}			
		}

		if (header.CHRROM_count > 0) {
			// Allocate a pointer for each chrrom page
			CHRROM_Pages = new Uint8 *[header.CHRROM_count];
							
			// Loop thorugh all pages
			int block_ctr = 0;
			while (block_ctr < header.CHRROM_count) {
				// Allocate storage for each page
				CHRROM_Pages[block_ctr] = new Uint8[8192];

				// Load each page
				memcpy(&CHRROM_Pages[block_ctr][0], &cartdata[16 + header.PRGROM_count*16384 + block_ctr*8192], 8192);
				block_ctr++;
			}
		}			
	}	
}



//////////////////////////////////////////////////////////////////////
// Procedure: NesCartridge Deconstructor
//////////////////////////////////////////////////////////////////////
// Description: The NesCartridge Deconstructor is in charge of deleting
// all of the PRGROM and CHRROM pages of memory loaded from a cartridge.
//////////////////////////////////////////////////////////////////////
NesCartridge::~NesCartridge()
{
	if (mintErrorCode == OK) // only free memory if the cartridge was actually loaded
	{
		int ctr = 0;
		while (ctr < header.PRGROM_count)
		{		
			if (PRGROM_Pages[ctr] != NULL)
			{
				delete[] PRGROM_Pages[ctr];			
			}
			ctr++;
		}
		if (PRGROM_Pages != NULL) delete[] PRGROM_Pages;
		PRGROM_Pages = NULL;

		ctr = 0;
		while (ctr < header.CHRROM_count)
		{
			if (CHRROM_Pages[ctr] != NULL)
			{
				delete[] CHRROM_Pages[ctr];			
			}
			ctr++;
		}
		if (CHRROM_Pages != 0) delete[] CHRROM_Pages;
		CHRROM_Pages = NULL;
	}	
	// PRGROM and CHRROM are only dynamically allocated memory in this class, done
}