//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// Nes Input.CPP - Implementation of the NesInput class.
//////////////////////////////////////////////////////////////////////

#include "NesInput.h"

//////////////////////////////////////////////////////////////////////
// Procedure: NesInput Constructor (special)
//////////////////////////////////////////////////////////////////////
// Description: This constructor initializes the controller and also
// designates itself as a given device.  This device assignment is probably
// not important as it is only used to give off the correct signature
// if a NES program should happen to check it, nothing else.
//////////////////////////////////////////////////////////////////////
NesInput::NesInput(int intNesDevice)
{
	mintErrorCode = 0x00;
	//clearStrobe();
	mintNesDevice = intNesDevice;
	bytController1Sig = 0x01;
	bytController2Sig = 0x02;
}



//////////////////////////////////////////////////////////////////////
// Procedure: NesInput Constructor
//////////////////////////////////////////////////////////////////////
// Description: This constructor initializes the controller, but assumes
// that the device it represents is the first controller.
//////////////////////////////////////////////////////////////////////
NesInput::NesInput()
{
	mintErrorCode = 0x00;
	bytStrobeLocation = 1;
	// Assume controller 1 if device not given
	mintNesDevice = CONTROLLER1;
	bytController1Sig = 0x01;
	bytController2Sig = 0x02;
}



//////////////////////////////////////////////////////////////////////
// Procedure: poll
//////////////////////////////////////////////////////////////////////
// Description: This procedure is called by a NES program repeatedly to
// get the button statuses.  HOWEVER, it is NOT where the physical devices
// (keyboard, etc) actually get polled!  This simply just shares the "buttonstatus"
// bit array with the NES!
//////////////////////////////////////////////////////////////////////
int NesInput::poll()
{
	int retval = 0;	
	if (bytStrobeLocation <= 0x08)
	{
		if (buttonStatus[bytStrobeLocation]) retval = 1;		
	} else if (bytStrobeLocation <= 12) {
		// Signature
		int intBitMask = 0x01 << (bytStrobeLocation - 9);		
		switch (mintNesDevice)
		{
		case CONTROLLER1:
			if (bytController1Sig & intBitMask) retval = 1;
			break;
		case CONTROLLER2:
			if (bytController2Sig & intBitMask) retval = 1;
			break;
		}
	} else {
		retval = 0;
	}
	bytStrobeLocation++;	
	if (bytStrobeLocation == 17) bytStrobeLocation = 1;
	return retval;
}


//////////////////////////////////////////////////////////////////////
// Procedure: clearStrobe
//////////////////////////////////////////////////////////////////////
// Description: This resets the buttonstatus bit array
//////////////////////////////////////////////////////////////////////
void NesInput::clearStrobe()
{
	int ctr= 1;
	while (ctr <= 14) {
		buttonStatus[ctr] = false;
		ctr++;
	}
	bytStrobeLocation = 1;
}



//////////////////////////////////////////////////////////////////////
// Procedure: resetStrobe
//////////////////////////////////////////////////////////////////////
// Description:  This procedure scans the input device, fills in the
// buttonstatus bit array, and sets the strobe location to 1.  
//
// IMPORTANT: This is the one (well, maybe two if you count the constructor)
// that a derived class will want to override!  This is where you check
// you keyboard, joypad, and fill in the buttonstatus array.  The base NesInput
// class will do the rest.
//////////////////////////////////////////////////////////////////////
void NesInput::resetStrobe() {	
	// dummy controller, don't mess with buttonstatus array
	bytStrobeLocation = 1;
}



//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the controller to the file stream
//-------------------------------------------------------------------
void NesInput::saveState(FILE *stream) {
	fwrite( &bytStrobeLocation, sizeof(unsigned char), 1, stream);	// Strobe location
	fwrite( &buttonStatus, sizeof(bool) * 9, 1, stream);			// Button status
}





//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the controller from the file stream
//-------------------------------------------------------------------
void NesInput::loadState(FILE *stream) {
	fread( &bytStrobeLocation, sizeof(unsigned char), 1, stream);	// Strobe location
	fread( &buttonStatus, sizeof(bool) * 9, 1, stream);				// Button status
}