//	UberNES Emulation Core
//  Copyright (C) 2004 Luke Molnar
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


//////////////////////////////////////////////////////////////////////
// Nes PPU Memory.cpp: implementation of the NesPPUMemory class
//////////////////////////////////////////////////////////////////////

#include "NesPPUMemory.h"

//////////////////////////////////////////////////////////////////////
// Procedure: NesPPUMemory constructor
//////////////////////////////////////////////////////////////////////
// Description:  Initializes the memory pointers to NULL
//////////////////////////////////////////////////////////////////////
NesPPUMemory::NesPPUMemory()
{
	memory = NULL;
	bytNameTable = NULL;
	palette = NULL;
}


//////////////////////////////////////////////////////////////////////
// Procedure: NesPPUMemory Deconstructor
//////////////////////////////////////////////////////////////////////
// Description: The NesPPUMemory Deconstructor simply deletes all memory
// associated with VRAM.
//////////////////////////////////////////////////////////////////////
NesPPUMemory::~NesPPUMemory()
{
	if (memory != NULL) delete memory;
	delete[] bytNameTable[0];
	delete[] bytNameTable[1];
	delete[] bytNameTable[2];
	delete[] bytNameTable[3];
	delete[] bytNameTable;
	delete palette;
}



//////////////////////////////////////////////////////////////////////
// Procedure: init
//////////////////////////////////////////////////////////////////////
// Description: Initializes the memory and sets some miscellaneous values
//////////////////////////////////////////////////////////////////////
void NesPPUMemory::init()
{
	if (memory != NULL) delete memory;
	memory = new Uint8[0x2000 + 1 ];

	int intCtr = 0;
	while (intCtr <= 0x2000) { memory[intCtr] = 0; intCtr++;}
	
	// Allocate two 1024 byte arrays for the two unique name tables the NES can store
	if (bytNameTable == NULL) {
		bytNameTable = new Uint8 *[4];
		bytNameTable[0] = new Uint8[0x400 + 1];
		bytNameTable[1] = new Uint8[0x400 + 1];
		bytNameTable[2] = new Uint8[0x400 + 1];
		bytNameTable[3] = new Uint8[0x400 + 1];
	}
	
	// Zero out the name table data
	int intNTCtr = 0;
	while (intNTCtr < 4) {
		int intCtr = 0;
		while (intCtr < 0x400) {
			bytNameTable[intNTCtr][intCtr] = 0;
			intCtr++;
		}
		intNTCtr++;
	}

	// Allocate and zero out palette memory
	if (palette != NULL) delete palette;
	palette = new Uint8[0x20 + 1];
	while (intCtr <= 0x20) { palette[intCtr] = 0; intCtr++; }

	mblnFirstRead = true;
	mbytReadBuffer = 0;
	mintIncrement = 1;
	mintSSMirrorLocation = 0x2000;
}


//////////////////////////////////////////////////////////////////////
// Procedure: read (internal PPU version)
//////////////////////////////////////////////////////////////////////
// Description:  This is the "read" method that is called by the PPU class.
// It is separate from the CPU version because the CPU version requires
// the address to be passed by reference, but it is more convenient for the
// PPU to read by value.  So this procedure basically just changes the 
// parameter from by value to by reference.
//////////////////////////////////////////////////////////////////////
Uint8 NesPPUMemory::read(Uint16 addr) {
	Uint16 tmpAddr = addr;
	return read(tmpAddr, false);
}




//////////////////////////////////////////////////////////////////////
// Procedure: read (for use by CPU/main read operation)
//////////////////////////////////////////////////////////////////////
// Description:  This procedure reads a byte of memory from the PPU memory.
// It is responsible for maintaining all of the various mirroring schemes
// the PPU can do, as well as the "buffered VRAM reading" quirk.
//
// If the second parameter, blnCPURead, is set to true, the address parameter
// will be auto-incremented (hence the pass-by-reference).  blnCPURead
// should only be set to true when the PRG-ROM issues a lda $2007 code.
//////////////////////////////////////////////////////////////////////
Uint8 NesPPUMemory::read(Uint16 &addr, bool blnCPURead)
{	
	// Only concerned with addresses $0 to $3FFF
	Uint16 tmpAddr = addr & 0x3FFF;

	// Store the address value in the last read latch
	mintLastReadAddr = addr;

	if ( (mblnFirstRead) && (blnCPURead) )
	{
		// We have to do some checking to see if this was the first read
		mblnFirstRead = false;		
		// (the first read is invalid so we read a value and then discard it)
		unsigned char temp = read(addr, blnCPURead);
		return 0;
	} else {	
		Uint8 bytRetVal = 0;
				
		if (tmpAddr < 0x2000)
		{
			// $0000 to $2000 = pattern tables
			bytRetVal =  memory[tmpAddr];
		} else {
			if (tmpAddr < 0x3F00)
			{
				// $2000 to $3000 is the name tables, while $3000 to $3F00 is a mirror of $2000 to $2F00
				tmpAddr &= 0x2FFF;

				// Return data from proper name table depending upon what mirror type is designated
				switch (mbytMirroringType)
				{				
				case VERTICAL:			// Vertical mirroring
					switch (tmpAddr & 0x0C00)
					{					
					case 0x000:	// NT #0, NT #3 point to the first logical name table
					case 0x800:	
						bytRetVal = bytNameTable[0][tmpAddr & 0x3FF];
						break;
					
					case 0x400:	// NT #1, NT #4 point to the second logical name table
					case 0xC00:
						bytRetVal = bytNameTable[1][tmpAddr & 0x3FF];
						break;
					}
					break;
								
				case HORIZONTAL:		// Horizontal mirroring
					switch (tmpAddr & 0x0C00)
					{					
					case 0x000:	// NT #0, NT #1 point to the first logical name table
					case 0x400:	
						bytRetVal = bytNameTable[0][tmpAddr & 0x3FF];
						break;
					
					case 0x800:	// NT #2, NT #3 point to the second logical name table
					case 0xC00:
						bytRetVal = bytNameTable[1][tmpAddr & 0x3FF];
						break;
					}
					break;
					
				case SINGLESCREEN:		// Single Screen mirroring
					if ( mintSSMirrorLocation == 0x2000 )
					{
						bytRetVal = bytNameTable[0][tmpAddr & 0x3FF];
					} else {
						bytRetVal = bytNameTable[1][tmpAddr & 0x3FF];
					}				
					break;
				}
			} else {
				// $3F00 to $4000 is the palette, mirrored 7 times
				tmpAddr &= 0x1F;

				if (tmpAddr & 0x03)
				{
					// If entry is not divisble by 4 then return palette value
					bytRetVal = palette[tmpAddr];
				} else {
					// Otherwise, every fourth palette entry is a mirror of entry 0
					bytRetVal = palette[0];										
				}					
			}
		}

		// Auto increment internal PPU address by either 1 or 32 after read depending on PPU status
		// only is called by CPU class (i.e. increment if NES ROM reads/writes, but not if PPU
		// method calls this)	
		// NOTE: also do the buffered input stuff here since this is only true when lda $2007 is hit
		if (blnCPURead) 
		{
			addr += mintIncrement;	
			unsigned char temp = mbytReadBuffer;		
			mbytReadBuffer = bytRetVal;
			return temp;
		} else {		
			return bytRetVal;
		}						
	}
}




//////////////////////////////////////////////////////////////////////
// Procedure: write (internal PPU version)
//////////////////////////////////////////////////////////////////////
// Description:  Like read, this is a simplified write procedure for
// internal use by the PPU class.
//////////////////////////////////////////////////////////////////////
void NesPPUMemory::write(Uint16 addr, Uint8 data)
{
	Uint16 tmpAddr = addr;
	write (tmpAddr, data, false);
}



//////////////////////////////////////////////////////////////////////
// Procedure: write (CPU version)
//////////////////////////////////////////////////////////////////////
// Description:  This procedure writes a byte to PPU memory, handling
// all of the mirroring schemes.  
//
// This procedure will alter the address parameter is the blnCPURead
// parameter is set to true.
//////////////////////////////////////////////////////////////////////
void NesPPUMemory::write(Uint16 &addr, Uint8 data, bool blnCPURead)
{
	Uint16 tmpAddr = addr & 0x3FFF;

	if (tmpAddr < 0x2000)
	{
		// $0000 to $2000 is the pattern tables
		memory[ tmpAddr ] = data;
	} else {
		if (tmpAddr < 0x3F00)
		{
			// $2000 to $3000 is the name tables, while $3000 to $3F00 is a 
			// mirror of $2000 to $2F00
			tmpAddr &= 0x2FFF;

			switch (mbytMirroringType)
			{
				case VERTICAL:			// Vertical mirroring
					switch (tmpAddr & 0x0C00)
					{					
					case 0x000:	// NT #0, NT #3 point to the first logical name table
					case 0x800:	
						bytNameTable[0][tmpAddr & 0x3FF] = data;
						break;
					
					case 0x400:	// NT #1, NT #4 point to the second logical name table
					case 0xC00:
						bytNameTable[1][tmpAddr & 0x3FF] = data;
						break;
					}
					break;

				case HORIZONTAL:		// Horizontal mirroring
					switch (tmpAddr & 0x0C00)
					{					
					case 0x000:	// NT #0, NT #1 point to the first logical name table
					case 0x400:	
						bytNameTable[0][tmpAddr & 0x3FF] = data;
						break;
					
					case 0x800:	// NT #2, NT #3 point to the second logical name table
					case 0xC00:
						bytNameTable[1][tmpAddr & 0x3FF] = data;
						break;
					}
					break;
					
				case SINGLESCREEN:		// Single Screen mirroring
					if ( mintSSMirrorLocation == 0x2000 )
					{
						bytNameTable[0][tmpAddr & 0x3FF] = data;
					} else {
						bytNameTable[1][tmpAddr & 0x3FF] = data;
					}				
					break;

			}			
		} else {
			// $3F00 to $4000 is the palette memory ($20 bytes mirrored 7 times)			
			tmpAddr &= 0x1F;
			data &= 0x3F;		// not sure if needed but there for completeness
			
			// Here's the deal with palette mirroing (within a $20 byte palette)
			// First of all, the palette data gets written to the given address:
			palette[tmpAddr] = data;
			// Then, if the palette entry is divisible by 4, the data gets mirrored
			// across to the other $10 byte palette.  (To use the nesdev terminology, it
			// gets mirrored from the sprite palette to the image palette and vice versa)
			if (!(tmpAddr & 0x3)) palette[tmpAddr ^ 0x10] = data;			
		}
	}
	
	// Auto increment internal PPU address by either 1 or 32 after write depending on PPU status
	if (blnCPURead) addr += mintIncrement;
}





//-------------------------------------------------------------------
// Procedure: saveState
//-------------------------------------------------------------------
// Description: Saves the state of the PPU VRAM to the file stream
//-------------------------------------------------------------------
void NesPPUMemory::saveState(FILE *stream) {
	fwrite( &memory[0], 0x2000, 1, stream);		 // Pattern tables
	fwrite( &bytNameTable[0][0], 0x400, 1, stream); // Name Table #0
	fwrite( &bytNameTable[1][0], 0x400, 1, stream); // Name Table #1
	fwrite( &bytNameTable[2][0], 0x400, 1, stream); // Name Table #2
	fwrite( &bytNameTable[3][0], 0x400, 1, stream); // Name Table #3
	fwrite( &palette[0], 0x20, 1, stream);			// Palette

	fwrite( &mintIncrement, sizeof(mintIncrement), 1, stream);				// Misc
	fwrite( &mblnFirstRead, sizeof(mblnFirstRead), 1, stream);				// Misc
	fwrite( &mbytReadBuffer, sizeof(mbytReadBuffer), 1, stream);			// Misc
	fwrite( &mbytMirroringType, sizeof(mbytMirroringType), 1, stream);		// Misc
	fwrite( &mintSSMirrorLocation, sizeof(mintSSMirrorLocation), 1, stream);// Misc
	fwrite( &mintLastReadAddr, sizeof(mintLastReadAddr), 1, stream);		// Misc
	
	fwrite( &mintLatchValue1, sizeof(mintLatchValue1), 1, stream);			// Misc
	fwrite( &mintLatchValue2, sizeof(mintLatchValue2), 1, stream);			// Misc
	fwrite( &mintLatchValue3, sizeof(mintLatchValue3), 1, stream);			// Misc
	fwrite( &mintLatchValue4, sizeof(mintLatchValue4), 1, stream);			// Misc
	fwrite( &mintLatchSelect1, sizeof(mintLatchSelect1), 1, stream);			// Misc
	fwrite( &mintLatchSelect2, sizeof(mintLatchSelect2), 1, stream);			// Misc
	fwrite( &mintLatchSelect3, sizeof(mintLatchSelect3), 1, stream);			// Misc
	fwrite( &mintLatchSelect4, sizeof(mintLatchSelect4), 1, stream);			// Misc
}





//-------------------------------------------------------------------
// Procedure: loadState
//-------------------------------------------------------------------
// Description: Loads the state of the PPU VRAM from the file stream
//-------------------------------------------------------------------
void NesPPUMemory::loadState(FILE *stream) {
	fread( &memory[0], 0x2000, 1, stream);		   // Pattern tables
	fread( &bytNameTable[0][0], 0x400, 1, stream); // Name Table #0
	fread( &bytNameTable[1][0], 0x400, 1, stream); // Name Table #1
	fread( &bytNameTable[2][0], 0x400, 1, stream); // Name Table #2
	fread( &bytNameTable[3][0], 0x400, 1, stream); // Name Table #3
	fread( &palette[0], 0x20, 1, stream);		   // Palette
	
	fread( &mintIncrement, sizeof(mintIncrement), 1, stream);				// Misc
	fread( &mblnFirstRead, sizeof(mblnFirstRead), 1, stream);				// Misc
	fread( &mbytReadBuffer, sizeof(mbytReadBuffer), 1, stream);				// Misc
	fread( &mbytMirroringType, sizeof(mbytMirroringType), 1, stream);		// Misc
	fread( &mintSSMirrorLocation, sizeof(mintSSMirrorLocation), 1, stream);	// Misc
	fread( &mintLastReadAddr, sizeof(mintLastReadAddr), 1, stream);			// Misc
	fread( &mintLatchValue1, sizeof(mintLatchValue1), 1, stream);			// Misc
	fread( &mintLatchValue2, sizeof(mintLatchValue2), 1, stream);			// Misc
	fread( &mintLatchValue3, sizeof(mintLatchValue3), 1, stream);			// Misc
	fread( &mintLatchValue4, sizeof(mintLatchValue4), 1, stream);			// Misc
	fread( &mintLatchSelect1, sizeof(mintLatchSelect1), 1, stream);			// Misc
	fread( &mintLatchSelect2, sizeof(mintLatchSelect2), 1, stream);			// Misc
	fread( &mintLatchSelect3, sizeof(mintLatchSelect3), 1, stream);			// Misc
	fread( &mintLatchSelect4, sizeof(mintLatchSelect4), 1, stream);			// Misc	
}


